classdef STFT
    % STFT Short-Time Fourier Transform
    
    properties
        windowLength; % window length
        windowHop; % window hop
        window; % window type
        fftLength; % FFT length
        samplerate; % sample rate
    end
    
    methods
        
        function stft = STFT(parWindowLength, parWindowHop, ...
                parWindowType, parFftLength, parSamplerate)
            % STFT Construct an instance of this class

            stft.windowLength = parWindowLength;
            if ~exist('parWindowHop','var') || isempty(parWindowHop)
                stft.windowHop = parWindowLength / 4; % default window hop
            else
                stft.windowHop = parWindowHop;
            end
            if ~exist('parWindowType','var') || isempty(parWindowType)
                stft.window = hamming(parWindowLength, 'periodic'); % default window type
            else
                if strncmpi(parWindowType, 'hamming', 1)
                    stft.window = hammingf(parWindowLength, 'periodic');
                elseif strncmpi(parWindowType, 'hann', 1)
                    stft.window = hannf(parWindowLength, 'periodic');
                elseif strncmpi(parWindowType, 'gauss', 1)
                    stft.window = gausswinf(parWindowLength);
                elseif strncmpi(parWindowType, 'tukey', 1)
                    stft.window = tukeywinf(parWindowLength);
                end
            end
            if ~exist('parFftLength','var') || isempty(parFftLength)
                stft.fftLength = 2^nextpow2(parWindowLength); % default fft length
            else
                stft.fftLength = parFftLength;
            end
            if ~exist('parSamplerate','var') || isempty(parSamplerate)
                stft.samplerate = 16000; % default samplerate
            else
                stft.samplerate = parSamplerate;
            end
        end
        
        function [X,t,f] = stft(stft, x)
            % STFT Perform short-time Fourier transform returns time slots
            % and number of frequency bins
            
            x = x(:);
            signalLength = length(x);
            
            numFreqbins = ceil((1+stft.fftLength)/2);
            numTimeslots = 1 + fix((signalLength-stft.windowLength)/stft.windowHop);
            
            % preallocate STFT array
            X = zeros(numFreqbins, numTimeslots);
            
            timeslotIndex = 0;
            for timeslot = 1 : numTimeslots
                % windowing
                xtemp = x(timeslotIndex+1:timeslotIndex+stft.windowLength);
                xtemp = xtemp .* stft.window;
                
                % FFT
                Xtemp = fft(xtemp, stft.fftLength);
                X(:,timeslot) = Xtemp(1:numFreqbins);
                
                % next hop
                timeslotIndex = timeslotIndex + stft.windowHop;
            end
            
            % time and frequency vectors for STFT
            t = (stft.windowLength/2 : stft.windowHop : ...
                stft.windowLength/2+(numTimeslots-1)*stft.windowHop) / stft.samplerate;
            f = (0:numFreqbins-1)*stft.samplerate/stft.fftLength;
            
        end % stft
        
        function [x,t] = istft(stft, X)
            % ISTFT Perform inverse short-time Fourier transform
            
            numTimeslots = size(X, 2);
            
            % preallocate ISTFT vector
            signalLength = stft.windowLength+(numTimeslots-1)*stft.windowHop;
            x = zeros(1, signalLength);
            
            timeslotIndex = 0;
            for timeslot = 1 : numTimeslots
                Xt = X(:, timeslot);
                if rem(stft.fftLength, 2)
                    % odd fftLength: excl. Nyquist frequency
                    Xtemp = [Xt; conj(Xt(end:-1:2))];
                else
                    % even fftLength: incl. Nyquist frequency
                    Xtemp = [Xt; conj(Xt(end-1:-1:2))];
                end
                
                % IFFT
                xtemp = real(ifft(Xtemp, stft.fftLength));
                xtemp = xtemp(1:stft.windowLength);
                
                % overlap-add method
                x((timeslotIndex+1):(timeslotIndex+stft.windowLength)) = ...
                    x((timeslotIndex+1):(timeslotIndex+stft.windowLength)) + (xtemp.*stft.window)';
                
                % next hop
                timeslotIndex = timeslotIndex + stft.windowHop;
            end
            
            % signal and time vector
            x = x.*stft.windowHop/sum(stft.window.^2);
            t = (0:signalLength-1)/stft.samplerate;
            
        end % istft
        
        function [fig] = plotTotal(stft, X, titlestr, maxFrequency)
            % PLOTTOTAL Plot entire STFT representation as color plot
            
            if ~exist('maxFrequency','var') || isempty(maxFrequency)
                maxFrequency = stft.samplerate / 2; % default top freq limit
            end
            numTimeslots = length(X(1, :));
            numFreqbins = length(X(:, 1));
            timeslots = (stft.windowLength/2 : stft.windowHop : ...
                stft.windowLength/2+(numTimeslots-1)*stft.windowHop) / stft.samplerate;
            freqbins = (0:numFreqbins-1)*stft.samplerate/stft.fftLength;
            
            fig = figure;
            %resizeFigure(fig, 800, 600);
            
            mesh(timeslots, freqbins, 20*log10(abs(X)));
            view(2);
            xlim([0 timeslots(end)]);
            ylim([0 maxFrequency]);
            caxis([-160 20]);
            colorbar;
            xlabel('Time (s)');
            ylabel('Frequency (Hz)');
            zlabel('Mag (dB)');
            title([titlestr,': Magnitude (dB)']);
            
        end % plotTotal
        
        function [fig] = plotFreqbins(stft, X, freqindecies, titlestr)
            % PLOTFREQBINS Plot STFT representation over frequency bins

            numFreqindecies = length(freqindecies);
            if numFreqindecies > 12, warning('Trying to plot more than 12 frequency bins.'); end
            numTimeslots = length(X(1, :));
            numFreqbins = length(X(:, 1));
            timeslots = (stft.windowLength/2 : stft.windowHop : ...
                stft.windowLength/2+(numTimeslots-1)*stft.windowHop) / stft.samplerate;
            freqbins = (0:numFreqbins-1)*stft.samplerate/stft.fftLength;

            pad = stft.samplerate/stft.fftLength/2;

            fig = figure;
            resizeFigure(fig, 600, 1000);

            for i = 1 : numFreqindecies
                freqindex = freqindecies(i);
                freqbin = freqbins(freqindex);
                subplot(numFreqindecies, 1, i);
                plot(timeslots, 20*log10(abs(X(freqindex, :))), 'k');
                ylim([-80,40]);
                if(i==1), title([titlestr,': Magnitude (dB) over Time (s)']); end
                legend(sprintf('%dHz < f < %dHz', round(freqbin-pad), round(freqbin+pad)));
            end

        end % plotFreqbins
        
        function [fig] = plotTimeslots(stft, X, timeindecies, titlestr)
            % PLOTTIMESLOTS Plot STFT representation over timeslots

            numTimeindecies = length(timeindecies);
            if numTimeindecies > 12, warning('Trying to plot more than 12 timeslots.'); end
            numTimeslots = length(X(1, :));
            numFreqbins = length(X(:, 1));
            timeslots = (stft.windowLength/2 : stft.windowHop : ...
                stft.windowLength/2+(numTimeslots-1)*stft.windowHop) / stft.samplerate;
            freqbins = (0:numFreqbins-1)*stft.samplerate/stft.fftLength;

            pad = 1000*stft.windowLength/stft.samplerate/2;

            fig = figure;
            resizeFigure(fig, 600, 1000);

            for i = 1 : numTimeindecies
                timeindex = timeindecies(i);
                timeslot = 1000*timeslots(timeindex);
                subplot(numTimeindecies, 1, i);
                plot(freqbins, 20*log10(abs(X(:, timeindex))), 'k');
                ylim([-80,40]);
                if(i==1), title([titlestr,': Magnitude (dB) over Frequency (Hz)']); end
                legend(sprintf('%dms < t < %dms', timeslot-pad, timeslot+pad));
            end

        end % plotTimeslots
        
    end % methods
    
end % classdef
